@@ -91,8 +91,8 @@ class getter:
         """
 
         value = self.info.get(key, key)
-        if isinstance(value, list):
-            value = value[self.index]
+        # if isinstance(value, list):
+        #     value = value[self.index]
         if isinstance(value, str):
             value = value.format(**self.info)
             value = self.header.get(value, alt)
@@ -133,8 +133,8 @@ class Instrument:
         self.find_closest = [
             "bias",
             "flat",
-            "wavecal",
-            "comb",
+            "wavecal_master",
+            "freq_comb_master",
             "orders",
             "scatter",
             "curvature",
@@ -143,6 +143,11 @@ class Instrument:
     def __str__(self):
         return self.name
 
+    def get(self, key, header, mode, alt=None):
+        get = getter(header, self.info, mode)
+        return get(key, alt=alt)
+
+
     def get_extension(self, header, mode):
         mode = mode.upper()
         extension = self.info.get("extension", 0)
@@ -269,6 +274,9 @@ class Instrument:
             jd = jd.to_value("mjd")
 
         header["e_orient"] = get("orientation", 0)
+        # As per IDL rotate if orient is 4 or larger and transpose is undefined
+        # the image is transposed
+        header["e_transpose"] = get("transpose", (header["e_orient"] % 8 >= 4))
 
         naxis_x = get("naxis_x", 0)
         naxis_y = get("naxis_y", 0)
@@ -302,9 +310,10 @@ class Instrument:
         header["e_obslat"] = get("latitude")
         header["e_obsalt"] = get("altitude")
 
-        header["HIERARCH e_wavecal_element"] = get(
-            "wavecal_element", info.get("wavecal_element", None)
-        )
+        if info.get("wavecal_element", None) is not None:
+            header["HIERARCH e_wavecal_element"] = get(
+                "wavecal_element", info.get("wavecal_element", None)
+            )
         return header
 
     def find_files(self, input_dir):
@@ -352,12 +361,12 @@ class Instrument:
                 "night": night,
                 "curvature": self.info["id_curvature"],
             },
-            "wavecal": {
+            "wavecal_master": {
                 "instrument": self.info["id_instrument"],
                 "night": night,
                 "wave": self.info["id_wave"],
             },
-            "freq_comb": {
+            "freq_comb_master": {
                 "instrument": self.info["id_instrument"],
                 "night": night,
                 "comb": self.info["id_comb"],
@@ -581,6 +590,8 @@ class Instrument:
         fname = os.path.join(cwd, "..", "masks", fname)
         return fname
 
+    def get_wavelength_range(self, header, mode, **kwargs):
+        return self.get("wavelength_range", header, mode)
 
 class InstrumentWithModes(Instrument):
     def __init__(self):
